#!/usr/bin/env python
# coding: utf-8

import logging
import json
import numpy as np
from collections import defaultdict

import amulog.manager
from amulog import common
from amulog import lt_common
from amulog.eval import cluster_metrics
from amulog.eval import structure_metrics

_logger = logging.getLogger(__package__.partition(".")[0])


class MeasureLTGen:
    """Record and load measurement data of tempalte generation.
    This is implemented in memory-saving architecture;
    The calculated data for each line is appended into text file."""

    SPLITTER = "@@"
    LABEL_NONE = "N"
    LABEL_DESCRIPTION = "D"
    LABEL_VARIABLE = "V"
    FILEPATH_DIGIT_LENGTH = 2

    def __init__(self, conf, n_trial):
        self._conf = conf
        self._n_trial = n_trial

        self._current_trial = None
        self._d_answer = None
        self._d_trial = None

    def _init_answer_info(self):
        common.mkdir(self._output_dir_answer(self._conf))
        self._d_answer = {"l_tid": list(),
                          "n_lines": int(),
                          "d_n_lines": defaultdict(int),
                          "n_words": int(),
                          "d_n_words": defaultdict(int),
                          }

    def _init_trial_info(self):
        common.mkdir(self._output_dir_trial(self._conf))
        if self._n_trial is None:
            return
        assert self._n_trial < 10 ** self.FILEPATH_DIGIT_LENGTH
        self._d_trial = {"l_tid": list(),
                         "n_c_lines": int(),
                         "d_n_c_lines": defaultdict(int),
                         "n_c_words": int(),
                         "d_n_c_words": defaultdict(int),
                         }

    # file IO methods
    @staticmethod
    def _output_dir_answer(conf):
        return conf["eval"]["ltgen_answer_dir"]

    @staticmethod
    def _output_dir_trial(conf):
        return conf["eval"]["ltgen_trial_dir"]

    def _org_word_path(self):
        return "{0}/word".format(self._output_dir_answer(self._conf))

    def _answer_label_path(self):
        return "{0}/label_answer".format(self._output_dir_answer(self._conf))

    def _trial_label_path(self, trial_id):
        str_trial_id = str(trial_id).zfill(self.FILEPATH_DIGIT_LENGTH)
        return "{0}/label_trial{1}".format(self._output_dir_trial(self._conf),
                                           str_trial_id)

    def _answer_info_path(self):
        return "{0}/info_answer".format(self._output_dir_answer(self._conf))

    def _trial_info_path(self, trial_id):
        str_trial_id = str(trial_id).zfill(self.FILEPATH_DIGIT_LENGTH)
        return "{0}/info_trial{1}".format(self._output_dir_trial(self._conf),
                                          str_trial_id)

    def init_answer(self):
        common.rm(self._org_word_path())
        common.rm(self._answer_label_path())
        self._init_answer_info()

    def init_trial(self, trial_id):
        common.rm(self._trial_label_path(trial_id))
        self._current_trial = trial_id
        self._init_trial_info()

    def _load_answer_info(self):
        with open(self._answer_info_path(), 'r', encoding='utf-8') as f:
            obj = json.load(f)
        self._d_answer = obj

    def _load_trial_info(self):
        trial_id = self._current_trial
        with open(self._trial_info_path(trial_id), 'r', encoding='utf-8') as f:
            obj = json.load(f)
        self._d_trial = obj

    def _dump_answer_info(self):
        obj = self._d_answer
        with open(self._answer_info_path(), 'w', encoding='utf-8') as f:
            json.dump(obj, f)

    def _dump_trial_info(self):
        trial_id = self._current_trial
        obj = self._d_trial
        with open(self._trial_info_path(trial_id), 'w', encoding='utf-8') as f:
            json.dump(obj, f)

    def load(self, trial_id=None):
        self._load_answer_info()
        if trial_id is not None:
            self.load_trial(trial_id)

    def load_trial(self, trial_id):
        self._current_trial = trial_id
        self._load_trial_info()

    def dump_answer(self):
        self._dump_answer_info()

    def dump_trial(self):
        self._dump_trial_info()

    # data APIs
    @classmethod
    def _tpl2dump(cls, tpl, l_w):
        if tpl is None:
            return cls.LABEL_NONE + "\n"
        else:
            l_label = [cls.LABEL_VARIABLE if w_tpl == lt_common.REPLACER
                       else cls.LABEL_DESCRIPTION
                       for w_tpl, w_org in zip(tpl, l_w)]
            return "".join(l_label) + "\n"

    @classmethod
    def _labels_isnone(cls, labels):
        return cls.LABEL_NONE in labels

    @classmethod
    def restore_tpl(cls, labels, l_w):
        if labels is None or labels == cls.LABEL_NONE:
            return None
        else:
            return [lt_common.REPLACER if label == cls.LABEL_VARIABLE
                    else w
                    for label, w in zip(labels, l_w)]

    @classmethod
    def restore_result(cls, labels, l_w):
        if labels is None or labels == cls.LABEL_NONE:
            return None
        else:
            return [lt_common.REPLACER_HEAD + w + lt_common.REPLACER_TAIL
                    if label == cls.LABEL_VARIABLE else w
                    for label, w in zip(labels, l_w)]

    def add_org(self, l_w):
        added_line = self.SPLITTER.join(l_w) + "\n"
        with open(self._org_word_path(), 'a') as f:
            f.write(added_line)

    def add_answer(self, tid, tpl, l_w):
        self._update_stat_answer(tid, tpl)
        added_line = self._tpl2dump(tpl, l_w)
        with open(self._answer_label_path(), 'a') as f:
            f.write(added_line)

    def add_trial(self, tid_trial, tpl_trial,
                  tid_answer, tpl_answer, l_w):
        self._update_stat_trial(tid_trial, tpl_trial,
                                tid_answer, tpl_answer)
        added_line = self._tpl2dump(tpl_trial, l_w)
        with open(self._trial_label_path(self._current_trial), 'a') as f:
            f.write(added_line)

    def _update_stat_answer(self, tid, tpl):
        self._d_answer["l_tid"].append(tid)
        if tid is not None:
            self._d_answer["n_lines"] += 1
            self._d_answer["d_n_lines"][str(tid)] += 1
            n_words = len(tpl)
            self._d_answer["n_words"] += n_words
            self._d_answer["d_n_words"][str(tid)] += n_words

    def _update_stat_trial(self, tid_trial, tpl_trial,
                           tid_answer, tpl_answer):
        self._d_trial["l_tid"].append(tid_trial)
        if tid_answer is not None:
            if tpl_trial == tpl_answer:
                self._d_trial["n_c_lines"] += 1
                self._d_trial["d_n_c_lines"][str(tid_answer)] += 1

            assert len(tpl_trial) == len(tpl_answer)
            for w_trial, w_answer in zip(tpl_trial, tpl_answer):
                if w_trial == w_answer:
                    self._d_trial["n_c_words"] += 1
                    self._d_trial["d_n_c_words"][str(tid_answer)] += 1

    def iter_org(self):
        with open(self._org_word_path(), 'r') as f:
            for line in f:
                yield line.strip().split(self.SPLITTER)

    def iter_label_answer(self, pass_none=False):
        with open(self._answer_label_path(), 'r') as f:
            for line in f:
                labels_str = line.strip()
                if self._labels_isnone(labels_str):
                    if pass_none:
                        pass
                    else:
                        yield None
                else:
                    yield labels_str

    def iter_label_trial(self, pass_none=False):
        with open(self._trial_label_path(self._current_trial), 'r') as f:
            for line in f:
                labels_str = line.strip()
                if self._labels_isnone(labels_str):
                    if pass_none:
                        pass
                    else:
                        yield None
                else:
                    yield labels_str

    def iter_tpl_answer(self, pass_none=False, fill_wildcard=False):
        for l_w, labels in zip(self.iter_org(), self.iter_label_answer()):
            if labels is None:
                if pass_none:
                    pass
                else:
                    yield None
            elif fill_wildcard:
                yield self.restore_result(labels, l_w)
            else:
                yield self.restore_tpl(labels, l_w)

    def iter_tpl_trial(self, pass_none=False, fill_wildcard=False):
        for l_w, labels in zip(self.iter_org(), self.iter_label_trial()):
            if labels is None:
                if pass_none:
                    pass
                else:
                    yield None
            elif fill_wildcard:
                yield self.restore_result(labels, l_w)
            else:
                yield self.restore_tpl(labels, l_w)

    def tid_list_answer(self, pass_none=False):
        if pass_none:
            return [tid for tid in self._d_answer["l_tid"]
                    if tid is not None]
        else:
            return self._d_answer["l_tid"]

    def tid_list_trial(self, pass_none=False):
        if pass_none:
            return [tid for tid in self._d_trial["l_tid"]
                    if tid is not None]
        else:
            return self._d_trial["l_tid"]

    def valid_tid_list_answer(self):
        return self.tid_list_answer(pass_none=True)

    def valid_tid_list_trial(self):
        return self.tid_list_trial(pass_none=True)

    def iter_cluster_answer(self):
        return self._d_answer["n_lines"].keys().__iter__()

    def iter_cluster_trial(self):
        return np.unique(self.tid_list_trial(pass_none=True))

    def number_of_trials(self):
        return self._n_trial

    def number_of_messages(self):
        return self._d_answer["n_lines"]

    def number_of_answer_clusters(self):
        return len(self._d_answer["d_n_lines"])

    def number_of_answer_cluster_members(self):
        d = {}
        for key, val in self._d_answer["d_n_lines"].items():
            tid = int(key)
            d[tid] = val
        return d

    def number_of_trial_clusters(self):
        return np.unique(self.valid_tid_list_trial()).shape[0]

    # accuracy methods
    def word_accuracy(self, recalculation=False):
        if recalculation:
            iterable_tpl_answer = self.iter_tpl_answer(pass_none=True)
            iterable_tpl_trial = self.iter_tpl_trial(pass_none=True)
            return structure_metrics.word_accuracy(
                iterable_tpl_answer, iterable_tpl_trial)
        else:
            # n_words: Number of all words in dataset
            n_words = self._d_answer["n_words"]
            # n_c_words: Number of words correctly labeled in dataset
            n_c_words = self._d_trial["n_c_words"]
            return 1.0 * n_c_words / n_words

    def line_accuracy(self, recalculation=False):
        if recalculation:
            iterable_tpl_answer = self.iter_tpl_answer(pass_none=True)
            iterable_tpl_trial = self.iter_tpl_trial(pass_none=True)
            return structure_metrics.line_accuracy(
                iterable_tpl_answer, iterable_tpl_trial)
        else:
            # n_lines: Number of all lines in dataset
            n_lines = self._d_answer["n_lines"]
            # n_c_lines: Number of lines correctly labeled in dataset
            n_c_lines = self._d_trial["n_c_lines"]
            return 1.0 * n_c_lines / n_lines

    def tpl_word_accuracy(self, recalculation=False):
        if recalculation:
            iterable_tpl_answer = self.iter_tpl_answer(pass_none=True)
            iterable_tpl_trial = self.iter_tpl_trial(pass_none=True)
            l_tid_answer = self.tid_list_answer(pass_none=True)
            return structure_metrics.tpl_word_accuracy(
                iterable_tpl_answer, iterable_tpl_trial, l_tid_answer)
        else:
            # d_n_words: Number of words in a template cluster
            d_n_words = self._d_answer["d_n_words"]
            # d_n_c_words: Number of words correctly labeled in a template cluster
            d_n_c_words = self._d_trial["d_n_c_words"]

            l_acc = []
            for key in d_n_words:  # key: str(tid)
                l_acc.append(d_n_c_words.get(key, 0) / d_n_words.get(key, 0))
            return np.average(l_acc)

    def tpl_accuracy(self, recalculation=False):
        if recalculation:
            iterable_tpl_answer = self.iter_tpl_answer(pass_none=True)
            iterable_tpl_trial = self.iter_tpl_trial(pass_none=True)
            l_tid_answer = self.tid_list_answer(pass_none=True)
            return structure_metrics.tpl_accuracy(
                iterable_tpl_answer, iterable_tpl_trial, l_tid_answer)
        else:
            # d_n_lines: Number of lines in a template cluster
            d_n_lines = self._d_answer["d_n_lines"]
            # d_n_c_lines: Number of lines correctly labeled in a template cluster
            d_n_c_lines = self._d_trial["d_n_c_lines"]

            l_acc = []
            for key in d_n_lines:
                l_acc.append(d_n_c_lines.get(key, 0) / d_n_lines.get(key, 0))
            return np.average(l_acc)

    def tpl_word_accuracy_dist(self):
        # d_n_words: Number of words in a template cluster
        d_n_words = self._d_answer["d_n_words"]
        # d_n_c_words: Number of words correctly labeled in a template cluster
        d_n_c_words = self._d_trial["d_n_c_words"]

        ret = {}
        for key in d_n_words:  # key: str(tid)
            tid = int(key)
            ret[tid] = d_n_c_words.get(key, 0) / d_n_words.get(key, 0)
        return ret

    def tpl_line_accuracy_dist(self):
        # d_n_lines: Number of lines in a template cluster
        d_n_lines = self._d_answer["d_n_lines"]
        # d_n_c_lines: Number of lines correctly labeled in a template cluster
        d_n_c_lines = self._d_trial["d_n_c_lines"]

        ret = {}
        for key in d_n_lines:  # key: str(tid)
            tid = int(key)
            ret[tid] = d_n_c_lines.get(key, 0) / d_n_lines.get(key, 0)
        return ret

    def tpl_description_accuracy(self):
        iterable_tpl_answer = self.iter_tpl_answer(pass_none=True)
        iterable_tpl_trial = self.iter_tpl_trial(pass_none=True)
        l_tid_answer = self.tid_list_answer(pass_none=True)
        return structure_metrics.tpl_desc_accuracy(
            iterable_tpl_answer, iterable_tpl_trial, l_tid_answer)

    def tpl_variable_accuracy(self):
        iterable_tpl_answer = self.iter_tpl_answer(pass_none=True)
        iterable_tpl_trial = self.iter_tpl_trial(pass_none=True)
        l_tid_answer = self.tid_list_answer(pass_none=True)
        return structure_metrics.tpl_var_accuracy(
            iterable_tpl_answer, iterable_tpl_trial, l_tid_answer)

    def rand_score(self):
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return cluster_metrics.rand_score(l_tid_answer, l_tid_trial)

    def adjusted_rand_score(self):
        from sklearn.metrics import adjusted_rand_score as score
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return score(l_tid_answer, l_tid_trial)

    def f1_score(self):
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return cluster_metrics.precision_recall_fscore(
            l_tid_answer, l_tid_trial)[2]

    def parsing_accuracy(self):
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return cluster_metrics.parsing_accuracy(l_tid_answer, l_tid_trial)

    def cluster_accuracy(self):
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return cluster_metrics.cluster_accuracy(l_tid_answer, l_tid_trial)

    def overdiv_ratio(self):
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return cluster_metrics.over_division_cluster_ratio(l_tid_answer,
                                                           l_tid_trial)

    def overagg_ratio(self):
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return cluster_metrics.over_aggregation_cluster_ratio(
            l_tid_answer, l_tid_trial)

    def homogeneity_score(self):
        from sklearn.metrics import homogeneity_score as score
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return score(l_tid_answer, l_tid_trial)

    def completeness_score(self):
        from sklearn.metrics import completeness_score as score
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return score(l_tid_answer, l_tid_trial)

    def v_measure_score(self, beta=1.0):
        from sklearn.metrics import v_measure_score as score
        l_tid_answer = self.valid_tid_list_answer()
        l_tid_trial = self.valid_tid_list_trial()
        return score(l_tid_answer, l_tid_trial, beta=beta)


def measure_accuracy_answer(conf, targets, n_trial=None):
    timer = common.Timer("measure-accuracy answer", output=_logger)
    timer.start()
    mlt = MeasureLTGen(conf, n_trial)
    mlt.init_answer()

    from amulog import lt_import
    table_answer = lt_common.TemplateTable()
    ltgen_answer = lt_import.init_ltgen_import(conf, table_answer)

    for pline in amulog.manager.iter_plines(conf, targets):
        tid, _ = ltgen_answer.process_line(pline)
        if tid is None:
            tpl = None
        else:
            tpl = ltgen_answer.get_tpl(tid)
        mlt.add_org(pline["words"])
        mlt.add_answer(tid, tpl, pline["words"])
    mlt.dump_answer()
    timer.stop()
    return mlt


def measure_accuracy_trial_offline(conf, targets, n_trial=None, mlt=None):
    if n_trial is None:
        n_trial = int(conf["eval"]["n_trial"])
    if mlt is None:
        mlt = MeasureLTGen(conf, n_trial)
        mlt.load()

    for trial_id in range(n_trial):
        timer = common.Timer("measure-accuracy-offline trial{0}".format(
            trial_id), output=_logger)
        timer.start()
        mlt.init_trial(trial_id)
        table = lt_common.TemplateTable()
        ltgen = amulog.manager.init_ltgen_methods(conf, table)

        input_lines = list(amulog.manager.iter_plines(conf, targets))
        d_plines = {mid: pline for mid, pline in enumerate(input_lines)}
        d_tid = ltgen.process_offline(d_plines)

        iterobj = zip(input_lines,
                      mlt.tid_list_answer(),
                      mlt.iter_tpl_answer())
        for mid, (pline, tid_answer, tpl_answer) in enumerate(iterobj):
            if tid_answer is None:
                tid_trial = None
                tpl_trial = None
            else:
                tid_trial = d_tid[mid]
                if tid_trial is None:
                    tpl_trial = None
                else:
                    tpl_trial = ltgen.get_tpl(tid_trial)
            mlt.add_trial(tid_trial, tpl_trial,
                          tid_answer, tpl_answer, pline["words"])
        mlt.dump_trial()
        timer.stop()

    return mlt


def measure_accuracy_trial_online(conf, targets_train, targets_test,
                                  n_trial=None, mlt=None):
    if n_trial is None:
        n_trial = int(conf["eval"]["n_trial"])
    if mlt is None:
        mlt = MeasureLTGen(conf, n_trial)
        mlt.load()

    from amulog import log_db
    for trial_id in range(n_trial):
        timer = common.Timer("measure-accuracy-online trial{0}".format(
            trial_id), output=_logger)
        timer.start()
        mlt.init_trial(trial_id)
        table = lt_common.TemplateTable()
        ltgen = amulog.manager.init_ltgen_methods(conf, table)

        if targets_train is not None:
            iterobj = amulog.manager.iter_plines(conf, targets_train)
            d_plines = {mid: pline for mid, pline in enumerate(iterobj)}
            ltgen.process_offline(d_plines)

        iterobj = zip(amulog.manager.iter_plines(conf, targets_test),
                      mlt.tid_list_answer(),
                      mlt.iter_tpl_answer())
        for pline, tid_answer, tpl_answer in iterobj:
            if tid_answer is None:
                tid_trial = None
                tpl_trial = None
            else:
                tid_trial, _ = ltgen.process_line(pline)
                tpl_trial = ltgen.get_tpl(tid_trial)
            mlt.add_trial(tid_trial, tpl_trial,
                          tid_answer, tpl_answer, pline["words"])
        mlt.dump_trial()
        timer.stop()

    return mlt


def get_accuracy_average(conf, n_trial, functions):
    mlt = MeasureLTGen(conf, n_trial)
    mlt.load()

    results = []
    for trial_id in range(n_trial):
        mlt.load_trial(trial_id)
        d_values = {}
        for func_name in functions:
            d_values[func_name] = getattr(mlt, func_name)()
        results.append(d_values)

    d_average = {func_name: np.average([d_values[func_name] for d_values in results])
                 for func_name in functions}
    return d_average


def get_templates(conf, n_trial, trial_id=0, answer=False, mlt=None):
    """Get template list after all log parsing.
    In online algorithms, template structure can be changed while processing.
    This function pick up a result for the last message with each template.
    """
    if mlt is None:
        mlt = MeasureLTGen(conf, n_trial)
        mlt.load(trial_id)

    if answer:
        tids = np.array(mlt.tid_list_answer())
        iterobj = mlt.iter_tpl_answer()
    else:
        tids = np.array(mlt.tid_list_trial())
        iterobj = mlt.iter_tpl_trial()

    d_last_index = defaultdict(int)
    for mid, tid in enumerate(tids):
        if tid is not None:
            d_last_index[tid] = mid
    d_last_index_rev = {mid: tid for tid, mid in d_last_index.items()}

    d_tpl = {}
    for mid, tpl in enumerate(iterobj):
        if mid in d_last_index_rev:
            tid = d_last_index_rev[mid]
            d_tpl[tid] = tpl

    return d_tpl


def offline_structure_metrics(conf, n_trial, trial_id=0, partial=False):
    mlt = MeasureLTGen(conf, n_trial)
    mlt.load(trial_id)

    d_tpl = get_templates(conf, n_trial, trial_id, mlt=mlt)
    tids = mlt.tid_list_trial(pass_none=True)
    word_acc = structure_metrics.word_accuracy(
        mlt.iter_tpl_answer(pass_none=True),
        map(lambda x: d_tpl[x], tids))
    line_acc = structure_metrics.line_accuracy(
        mlt.iter_tpl_answer(pass_none=True),
        map(lambda x: d_tpl[x], tids))
    tpl_acc = structure_metrics.tpl_accuracy(
        mlt.iter_tpl_answer(pass_none=True),
        map(lambda x: d_tpl[x], tids), tids)
    tpl_word_acc = structure_metrics.tpl_word_accuracy(
        mlt.iter_tpl_answer(pass_none=True),
        map(lambda x: d_tpl[x], tids), tids)

    if partial:
        tpl_desc_fail = structure_metrics.tpl_desc_accuracy(
            mlt.iter_tpl_answer(pass_none=True),
            map(lambda x: d_tpl[x], tids), tids)
        tpl_var_fail = structure_metrics.tpl_var_accuracy(
            mlt.iter_tpl_answer(pass_none=True),
            map(lambda x: d_tpl[x], tids), tids)
        ret = (word_acc, line_acc, tpl_acc, tpl_word_acc,
               tpl_desc_fail, tpl_var_fail)
        return ret
    else:
        return word_acc, line_acc, tpl_acc, tpl_word_acc


def search_fail_template(conf, n_trial, trial_id=0, pass_similar=True):
    mlt = MeasureLTGen(conf, n_trial)
    mlt.load(trial_id)

    s_pass = set()
    iterobj = zip(mlt.iter_org(), mlt.tid_list_answer(),
                  mlt.iter_label_answer(), mlt.iter_label_trial())
    for l_w, tid_answer, labels_answer, labels_trial in iterobj:
        if pass_similar and tid_answer in s_pass:
            continue
        if labels_answer == labels_trial:
            pass
        else:
            result_answer = mlt.restore_result(labels_answer, l_w)
            print("Answer: {0}".format(" ".join(result_answer)))
            result_trial = mlt.restore_result(labels_trial, l_w)
            print("Trial: {0}".format(" ".join(result_trial)))
            print("--------------------")
        pass
        s_pass.add(tid_answer)


def search_diff_template(conf1, conf2, n_trial,
                         trial_id1=0, trial_id2=0, pass_similar=True):
    mlt1 = MeasureLTGen(conf1, n_trial)
    mlt1.load(trial_id1)
    mlt2 = MeasureLTGen(conf2, n_trial)
    mlt2.load(trial_id2)

    s_pass = set()
    iterobj = zip(mlt1.iter_org(),
                  mlt1.tid_list_answer(),
                  mlt1.iter_label_answer(),
                  mlt1.iter_label_trial(),
                  mlt2.iter_label_trial())
    for l_w, tid_answer, labels_answer, labels_trial1, labels_trial2 in iterobj:
        if pass_similar and tid_answer in s_pass:
            continue
        if (not labels_trial1 == labels_answer) and \
                (labels_trial2 == labels_answer):
            tpl_answer = mlt1.restore_result(labels_answer, l_w)
            tpl_trial1 = mlt1.restore_result(labels_trial1, l_w)
            print("< Answer: {0}".format(" ".join(tpl_answer)))
            print("< Trial: {0}".format(" ".join(tpl_trial1)))
            print("--------------------")
        elif (labels_trial1 == labels_answer) and \
                (not labels_trial2 == labels_answer):
            tpl_answer = mlt1.restore_result(labels_answer, l_w)
            tpl_trial2 = mlt2.restore_result(labels_trial2, l_w)
            print("> Answer: {0}".format(" ".join(tpl_answer)))
            print("> Trial: {0}".format(" ".join(tpl_trial2)))
            print("--------------------")
        s_pass.add(tid_answer)


def _sample_partial_cluster(a_true, a_pred, n_samples):
    from sklearn.metrics.cluster import contingency_matrix
    cm = contingency_matrix(a_true, a_pred, sparse=True)
    # sklearn.metrics.cluster.contingency_matrix now uses
    # inverse output of np.unique(a_true, inverse=true)
    # as the input of contingency matrix.
    # Therefore, the unique output works as value mapping.
    a_true_map, a_true_inverse = np.unique(a_true, return_inverse=True)
    nz_true, _ = cm.nonzero()

    l_cluster = []
    for cls_true, uniq_cnt in zip(*np.unique(nz_true, return_counts=True)):
        if uniq_cnt > 1:
            tid_true = a_true_map[cls_true]
            div = []
            for tid_pred, cnt_pred in zip(*np.unique(
                    a_pred[a_true_inverse == cls_true], return_counts=True)):
                a_index = np.where((a_true == tid_true) &
                                   (a_pred == tid_pred))[0]
                tmp_n_samples = min(n_samples, a_index.shape[0])
                a_index_sample = a_index[:tmp_n_samples]
                div.append((tid_pred, cnt_pred, a_index_sample))
            l_cluster.append((tid_true, div))

    return l_cluster


def _get_complete_clusters(a_true, a_pred):
    from sklearn.metrics.cluster import contingency_matrix
    cm = contingency_matrix(a_true, a_pred, sparse=True)
    nz_true, _ = cm.nonzero()

    tids = []
    for tid_true, uniq_cnt in zip(*np.unique(nz_true, return_counts=True)):
        if uniq_cnt == 1:
            tids.append(tid_true)
    return tids


def search_fail_overdiv(conf, n_trial, trial_id=0, n_samples=1):
    """Search failed log clusters of over-division.
    e.g., 1 cls in answer ≡ 3 cls in trial"""

    timer = common.Timer("test fail_overdiv", output=_logger)
    timer.start()

    mlt = MeasureLTGen(conf, n_trial)
    mlt.load(trial_id)

    # overdiv cluster information
    a_tid_answer = np.array(mlt.valid_tid_list_answer())
    a_tid_trial = np.array(mlt.valid_tid_list_trial())
    l_cluster = _sample_partial_cluster(a_tid_answer, a_tid_trial, n_samples)

    timer.lap("lap1")

    # make sample tpl list to show
    s_index_to_show = set()
    for _, div in l_cluster:
        samples = [a_index_sample for _, _, a_index_sample in div]
        s_index_to_show = s_index_to_show | set(np.ravel(samples))

    timer.lap("lap2")

    # get templates for the indexes to show
    iterobj = mlt.iter_tpl_trial(pass_none=True, fill_wildcard=True)
    d_result = {index: result for index, result in enumerate(iterobj)
                if index in s_index_to_show}

    timer.lap("lap3")

    # show
    for tid_answer, div in l_cluster:
        print("Template ID {0} (in answer)".format(tid_answer))
        iterobj = sorted(div, key=lambda x: x[1], reverse=True)
        for cls_id, (tid_trial, cnt_trial, a_index) in enumerate(iterobj):
            for index in a_index:
                print("{0} ({1}): {2}".format(cls_id, cnt_trial,
                                              " ".join(d_result[index])))
        print("--------------------")
    timer.stop()


def search_fail_overagg(conf, n_trial, trial_id=0, n_samples=1):
    """Search failed log clusters of over-division.
    e.g., 3 cls in answer ≡ 1 cls in trial"""

    mlt = MeasureLTGen(conf, n_trial)
    mlt.load(trial_id)

    # overagg cluster information
    a_tid_answer = np.array(mlt.valid_tid_list_answer())
    a_tid_trial = np.array(mlt.valid_tid_list_trial())
    l_cluster = _sample_partial_cluster(a_tid_trial, a_tid_answer, n_samples)

    # make sample tpl list to show
    s_index_to_show = set()
    for _, div in l_cluster:
        samples = [a_index_sample for _, _, a_index_sample in div]
        s_index_to_show = s_index_to_show | set(np.ravel(samples))

    # get templates for the indexes to show
    iterobj = mlt.iter_tpl_trial(pass_none=True, fill_wildcard=True)
    d_result = {index: result for index, result in enumerate(iterobj)
                if index in s_index_to_show}

    # show
    for tid_trial, div in l_cluster:
        print("Cluster {0} (in trial)".format(tid_trial))
        iterobj = sorted(div, key=lambda x: x[1], reverse=True)
        for tid_answer, cnt_answer, a_index in iterobj:
            for index in a_index:
                print("ltid {0} ({1}): {2}".format(tid_answer, cnt_answer,
                                                   " ".join(d_result[index])))
        print("--------------------")


def search_diff_overdiv(conf1, conf2, n_trial, trial_id=0, n_samples=1):
    """Search log clusters that is accurate in conf1,
    but failed of over-division in conf2.
    e.g., 1 cls in answer ≡ 1 cls in trial-conf1 ≡ 3 cls in trial-conf2"""

    mlt1 = MeasureLTGen(conf1, n_trial)
    mlt1.load(trial_id)
    mlt2 = MeasureLTGen(conf2, n_trial)
    mlt2.load(trial_id)

    # clusters accurate in conf1
    a_tid_answer = np.array(mlt1.valid_tid_list_answer())
    a_tid_trial1 = np.array(mlt1.valid_tid_list_trial())
    tids = _get_complete_clusters(a_tid_answer, a_tid_trial1)

    # cluster information that is overdiv in conf2
    a_tid_trial2 = np.array(mlt2.valid_tid_list_trial())
    l_cls_all = _sample_partial_cluster(a_tid_answer, a_tid_trial2, n_samples)
    l_cluster = [(tid_true, div) for tid_true, div
                 in l_cls_all if tid_true in tids]

    # make sample tpl list to show
    s_index_to_show = set()
    for _, div in l_cluster:
        samples = [a_index_sample for _, _, a_index_sample in div]
        s_index_to_show = s_index_to_show | set(np.ravel(samples))

    # get templates for the indexes to show
    iterobj = mlt2.iter_tpl_trial(pass_none=True, fill_wildcard=True)
    d_result = {index: result for index, result in enumerate(iterobj)
                if index in s_index_to_show}

    # show
    for tid_answer, div in l_cluster:
        print("Template ID {0} (in answer)".format(tid_answer))
        iterobj = sorted(div, key=lambda x: x[1], reverse=True)
        for cid, (tid_trial, cnt_trial, a_index) in enumerate(iterobj):
            for index in a_index:
                print("{0} ({1}): {2}".format(cid, cnt_trial,
                                              " ".join(d_result[index])))
        print("--------------------")


def search_diff_overagg(conf1, conf2, n_trial, trial_id=0, n_samples=1):
    """Search log clusters that is accurate in conf1,
    but failed of over-aggregation in conf2.
    e.g., 3 cls in answer ≡ 3 cls in trial-conf1 ≡ 1 cls in trial-conf2"""
    mlt1 = MeasureLTGen(conf1, n_trial)
    mlt1.load(trial_id)
    mlt2 = MeasureLTGen(conf2, n_trial)
    mlt2.load(trial_id)

    # clusters accurate in conf1
    a_tid_answer = np.array(mlt1.valid_tid_list_answer())
    a_tid_trial1 = np.array(mlt1.valid_tid_list_trial())
    tids = _get_complete_clusters(a_tid_answer, a_tid_trial1)

    # cluster information that is overagg in conf2
    a_tid_trial2 = np.array(mlt2.valid_tid_list_trial())
    l_cls_all = _sample_partial_cluster(a_tid_trial2, a_tid_answer, n_samples)
    l_cluster = []
    for tid_trial2, div in l_cls_all:
        for tid_answer, a_index, cnt in div:
            if tid_answer not in tids:
                break
        else:
            l_cluster.append((tid_trial2, div))

    # make sample tpl list to show
    s_index_to_show = set()
    for _, div in l_cluster:
        samples = [a_index_sample for _, _, a_index_sample in div]
        s_index_to_show = s_index_to_show | set(np.ravel(samples))

    # get templates for the indexes to show
    iterobj = mlt2.iter_tpl_trial(pass_none=True, fill_wildcard=True)
    d_result = {index: result for index, result in enumerate(iterobj)
                if index in s_index_to_show}

    # show
    for tid_trial2, div in l_cluster:
        print("Cluster {0} (in trial)".format(tid_trial2))
        iterobj = sorted(div, key=lambda x: x[1], reverse=True)
        for tid_answer, cnt_answer, a_index in iterobj:
            for index in a_index:
                print("ltid {0} ({1}): {2}".format(tid_answer, cnt_answer,
                                                   " ".join(d_result[index])))
        print("--------------------")


def measure_time_online(conf, targets_train, targets_test, n_trial=None):
    if n_trial is None:
        n_trial = int(conf["eval"]["n_trial_time"])

    d_time = {}
    for trial_id in range(n_trial):
        table = lt_common.TemplateTable()
        ltgen = amulog.manager.init_ltgen_methods(conf, table)
        if targets_train is not None:
            for pline in amulog.manager.iter_plines(conf, targets_train):
                ltgen.process_line(pline)

        timer = common.Timer("measure-time-online trial{0}".format(trial_id),
                             output=None)
        timer.start()
        for pline in amulog.manager.iter_plines(conf, targets_test):
            ltgen.process_line(pline)
        timer.stop()
        d_time[trial_id] = timer.total_time().total_seconds()

    return d_time


def measure_time_offline(conf, targets_test, n_trial=None):
    if n_trial is None:
        n_trial = int(conf["eval"]["n_trial_time"])

    d_time = {}
    for trial_id in range(n_trial):
        table = lt_common.TemplateTable()
        ltgen = amulog.manager.init_ltgen_methods(conf, table)

        timer = common.Timer("measure-time-offline trial{0}".format(trial_id),
                             output=None)
        timer.start()
        input_lines = list(amulog.manager.iter_plines(conf, targets_test))
        d_plines = {mid: pline for mid, pline in enumerate(input_lines)}
        ltgen.process_offline(d_plines)
        timer.stop()
        d_time[trial_id] = timer.total_time().total_seconds()

    return d_time



import os

from . import db_common


class Sqlite3(db_common.Database):

    def __init__(self, dbpath):
        self._dbpath = dbpath
        self._connect = None

    def __del__(self):
        if self._connect is not None:
            self._connect.commit()
            self._connect.close()

    def _open(self):
        import sqlite3 as sqlite3_mod
        self._connect = sqlite3_mod.connect(self._dbpath)
        self._connect.text_factory = str

    def db_exists(self):
        if os.path.exists(self._dbpath):
            return True
        else:
            return False

    def reset(self):
        if os.path.exists(self._dbpath):
            os.remove(self._dbpath)

    def commit(self):
        if self._connect is not None:
            self._connect.commit()

    def execute(self, sql, args=None):
        # print(sql)
        # if args is not None:
        #     print(args)
        if self._connect is None:
            self._open()
        cursor = self._connect.cursor()
        if args is None or len(args) == 0:
            cursor.execute(sql)
        else:
            cursor.execute(sql, args)
        return cursor

    def executemany(self, sql, iter_args):
        if self._connect is None:
            self._open()
        cursor = self._connect.cursor()
        cursor.executemany(sql, iter_args)
        return cursor

    def get_table_names(self):
        sql = "select name from sqlite_master"
        cursor = self.execute(sql)
        return [row[0] for row in cursor]

    def get_column_names(self, table_name):
        sql = ("select sql from sqlite_master "
               "where type='table' and name='{0}'".format(table_name))
        cursor = self.execute(sql)
        row = cursor.__iter__().__next__()
        table_sql = row[0]

        tmp_columns = table_sql.partition("(")[-1].rstrip(")")
        column_names = [part.strip(" ").split(" ")[0]
                        for part in tmp_columns.split(",")]
        return column_names

    # sql methods, basically classmethod or staticmethod
    @classmethod
    def datetime(cls, ret):
        return cls.strptime(ret)

    @staticmethod
    def _ph(varname):
        return ":{0}".format(varname)

    @staticmethod
    def _table_key_type(type_str):
        if type_str == "datetime":
            return "text"
        else:
            return type_str

    @staticmethod
    def _table_key_attr(attr):
        if attr == "primary_key":
            return "primary key"
        elif attr == "auto_increment":
            return "autoincrement"
        elif attr == "not_null":
            return "not null"
        else:
            raise NotImplementedError

    @staticmethod
    def _index_key(tablekey):
        return tablekey.key


def init_db_conn(conf):
    dbpath = conf.get("database", "sqlite3_filename")
    return Sqlite3(dbpath)



#!/usr/bin/env python

import sys
import os
import re
from setuptools import setup, find_packages


def load_readme():
    with open('README.rst', 'r') as fd:
        return fd.read()


def load_requirements():
    """Parse requirements.txt"""
    reqs_path = os.path.join('.', 'requirements.txt')
    with open(reqs_path, 'r') as fd:
        requirements = [line.rstrip() for line in fd]
    return requirements


sys.path.append("./tests")
package_name = 'amulog'
data_dir = "/".join((package_name, "data"))
data_files = ["/".join((data_dir, fn)) for fn in os.listdir(data_dir)]

init_path = os.path.join(os.path.dirname(__file__), package_name, '__init__.py')
with open(init_path) as f:
    version = re.search("__version__ = '([^']+)'", f.read()).group(1)

setup(name=package_name,
      version=version,
      description='A system log management tool with automatically generated log templates.',
      long_description=load_readme(),
      author='Satoru Kobayashi',
      author_email='sat@nii.ac.jp',
      url='https://github.com/cpflat/amulog/',
      install_requires=load_requirements(),
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Science/Research',
          "Intended Audience :: Developers",
          'License :: OSI Approved :: BSD License',
          "Operating System :: OS Independent",
          'Programming Language :: Python :: 3.6',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Software Development :: Libraries :: Python Modules'],
      license='The 3-Clause BSD License',

      packages=find_packages(),
      package_data={'amulog': data_files},
      include_package_data=True,
      entry_points={
          'console_scripts': [
              'amulog = amulog.__main__:main',
              'amulog.edit = amulog.edit.__main__:main',
              'amulog.eval = amulog.eval.__main__:main',
          ],
      },
      test_suite="tests"
      )



import datetime
from typing import NamedTuple, Iterable
from abc import ABC, abstractmethod


class TableKey(NamedTuple):
    """A namedtuple to define a column in CREATE syntax.
    Used in CREATE TABLE and CREATE INDEX.
    """
    key: str  #: column name
    type: str  #: type name, in SQL data types
    attr: Iterable  #: attribute of the column, e.g. autoincrement


class Condition(NamedTuple):
    """A namedtuple to specify a condition in WHERE syntax.
    Used in SELECT and UPDATE.

    If :attr:`repl` is True, :attr:`val` is considered as a placeholder.
    It will be replaced with the corresponding item in the "args" argument of :meth:`Database.execute`.
    """
    key: str  #: column name
    opr: str  #: operand to compare key and val
    val: str  #: compared value (or placeholder key if repl is True)
    repl: bool  #: If true, val is considered as a placeholder.


class StateSet(NamedTuple):
    """A namedtuple to specify a value in VALUES syntax.
    Used in INSERT and UPDATE.
    """
    key: str  #: column name
    val: str  #: assigned value


# to be removed
# from collections import namedtuple
# # namedtuple defaults is available after python 3.7
# # TableKey = namedtuple("TableKey", ("key", "type", "attr"), defaults=(tuple(),))
# TableKey = namedtuple("TableKey", ("key", "type", "attr"))
# Condition = namedtuple("Condition", ("key", "opr", "val", "repl"))
# StateSet = namedtuple("StateSet", ("key", "val"))


class Database(ABC):

    # d_key in create_table : key = key_name, val = [type, attr, attr...]
    # d_opr : operand to use in compararizon
    # l_repl : values with given keys are replaced in sql_query
    #          to generate sql with subquery

    # database management methods
    @abstractmethod
    def db_exists(self):
        raise NotImplementedError

    @abstractmethod
    def reset(self):
        raise NotImplementedError

    @abstractmethod
    def commit(self):
        raise NotImplementedError

    @abstractmethod
    def execute(self, sql, args):
        raise NotImplementedError

    @abstractmethod
    def executemany(self, sql, iter_args):
        raise NotImplementedError

    @abstractmethod
    def get_table_names(self):
        raise NotImplementedError

    def get_column_names(self, table_name):
        raise NotImplementedError

    # sql methods, basically classmethod or staticmethod
    @staticmethod
    def strftime(dt):
        if isinstance(dt, datetime.datetime):
            return dt.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(dt, str):
            return dt
        else:
            raise TypeError

    @staticmethod
    def strptime(string):
        return datetime.datetime.strptime(string, '%Y-%m-%d %H:%M:%S')

    @classmethod
    @abstractmethod
    def datetime(cls, ret):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _ph(varname):
        """place holder format"""
        raise NotImplementedError

    @classmethod
    def _set_state(cls, l_setstate):
        if len(l_setstate) == 0:
            raise ValueError("empty setstates")
        l_buf = []
        for ss in l_setstate:
            l_buf.append("{0} = {1}".format(ss.key, cls._ph(ss.val)))
        return ", ".join(l_buf)

    @classmethod
    def _cond_state(cls, l_cond):
        l_buf = []
        for cond in l_cond:
            if cond.repl:
                buf = "{0.key} {0.opr} {1}".format(cond, cls._ph(cond.val))
            else:
                buf = "{0.key} {0.opr} ({0.val})".format(cond)
            l_buf.append(buf)
        return " and ".join(l_buf)

    @staticmethod
    @abstractmethod
    def _table_key_type(type_str):
        # allowed typename : integer, text, datetime
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _table_key_attr(attr):
        # allowed attr : primary_key, auto_increment, not_null
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _index_key(tablekey):
        raise NotImplementedError

    @staticmethod
    def join_sql(join_opt, table_name1, table_name2, key1, key2):
        return "{1} {0} join {2} on {1}.{3} = {2}.{4}".format(
            join_opt, table_name1, table_name2, key1, key2)

    @classmethod
    def create_table_sql(cls, table_name, l_tablekey):
        l_def = []
        for key in l_tablekey:
            type_name = cls._table_key_type(key.type)
            if len(key.attr) > 0:
                l_attr = []
                for a in key.attr:
                    l_attr.append(cls._table_key_attr(a))
                l_def.append("{0} {1} {2}".format(key.key, type_name,
                                                  " ".join(l_attr)))
            else:
                l_def.append("{0} {1}".format(key.key, type_name))
        sql = "create table {0} ({1})".format(table_name, ", ".join(l_def))
        return sql

    @classmethod
    def create_index_sql(cls, table_name, index_name, l_tablekey):
        sql = "create index {0} on {1}({2})".format(
            index_name, table_name,
            ", ".join([cls._index_key(key) for key in l_tablekey])
        )
        return sql

    @classmethod
    def alter_table_sql(cls, old_table_name, new_table_name):
        sql = "alter table {0} rename to {1}".format(
            old_table_name, new_table_name
        )
        return sql

    @classmethod
    def select_sql(cls, table_name, l_key,
                   l_cond=None, l_order=None, opt=None):
        # now only "distinct" is allowed for opt
        sql_header = "select"
        if opt is not None and "distinct" in opt:
            sql_header += " distinct"
        sql = "{0} {1} from {2}".format(sql_header, ", ".join(l_key),
                                        table_name)
        if l_cond is not None and len(l_cond) > 0:
            sql += " where {0}".format(cls._cond_state(l_cond))
        if l_order is not None and len(l_order) > 0:
            sql_order = ", ".join(["{0} {1}".format(col, order)
                                   for col, order in l_order])
            sql += " order by " + sql_order
        return sql

    @classmethod
    def insert_sql(cls, table_name, l_setstate):
        l_key, l_val = zip(*[(ss.key, cls._ph(ss.val)) for ss in l_setstate])
        sql = "insert into {0} ({1}) values ({2})".format(
            table_name, ", ".join(l_key), ", ".join(l_val))
        return sql

    @classmethod
    def update_sql(cls, table_name, l_setstate, l_cond=None):
        sql = "update {0} set {1}".format(table_name,
                                          cls._set_state(l_setstate))
        if l_cond is not None and len(l_cond) > 0:
            sql += " where {0}".format(cls._cond_state(l_cond))
        return sql

    @classmethod
    def delete_sql(cls, table_name, l_cond=None):
        sql = "delete from {0}".format(table_name)
        if l_cond is not None and len(l_cond) > 0:
            sql += " where {0}".format(cls._cond_state(l_cond))
        return sql

    @staticmethod
    def drop_table_sql(table_name):
        return "drop table {0}".format(table_name)

    @staticmethod
    def drop_index_sql(index_name):
        return "drop index {0}".format(index_name)




"""argparse wrapper"""

import os
import sys
import argparse


def main(d_argset, d_alias=None, sublibs=None):
    command_path = os.path.basename(sys.argv[0])

    buf_usage = [
        ("usage: " + command_path + " SUBCOMMAND [options and arguments] ..."),
        "",
        "subcommands: "
    ]

    d_subc = {}
    for key, argset in d_argset.items():
        d_subc[key] = argset[0].partition("@")[0]
    if d_alias:
        for alias, key in d_alias.items():
            d_subc[alias] = "same as {0}".format(key)

    buf_usage += ["  {0}: {1}".format(k, v)
                  for k, v in sorted(d_subc.items())]

    buf_usage += [
        "",
        ("try \"" + command_path + " SUBCOMMAND -h\" "
         "to refer detailed subcommand usage")
    ]

    if sublibs:
        buf_usage.append(
            "see also sub-libraries {0}".format(
                " ".join(["amulog.{0}".format(n) for n in sublibs])
            )
        )

    usage = "\n".join(buf_usage)

    if d_alias:
        for alias, key in d_alias.items():
            d_argset[alias] = d_argset[key]

    if len(sys.argv) < 2:
        sys.exit(usage)
    mode = sys.argv[1]
    if mode in ("-h", "--help"):
        sys.exit(usage)
    commandline = sys.argv[2:]

    desc, l_argset, func = d_argset[mode]
    ap = argparse.ArgumentParser(prog=" ".join(sys.argv[0:2]),
                                 description=desc.replace("@", ""))
    for args, kwargs in l_argset:
        ap.add_argument(*args, **kwargs)
    ns = ap.parse_args(commandline)
    func(ns)




import os
import random
import logging
from collections import defaultdict

import amulog.manager
from amulog import config
from amulog import lt_common
from . import _items

_logger = logging.getLogger(__package__)


def _pop_fairly(d, size):
    ret = []
    keys = list(d.keys())
    random.shuffle(keys)
    while len(ret) < size:
        for key in keys:
            if len(d[key]) > 0:
                elm = d[key].pop()
                ret.append(elm)
            if len(ret) >= size:
                break
    return ret


def train_sample_random(l_lm, size):
    if len(l_lm) <= size:
        return l_lm
    else:
        return random.sample(l_lm, size)


def train_sample_ltgen(l_lm, size, ltgen):
    if len(l_lm) <= size:
        return l_lm

    plines = [{"words": lm.l_w, "symbols": lm.lt.lts} for lm in l_lm]
    d_ret = ltgen.process_offline(plines)
    d_lm = defaultdict(list)
    for mid, lm in enumerate(l_lm):
        tid = d_ret[mid]
        d_lm[tid].append(lm)
    for tid in d_lm:
        random.shuffle(d_lm[tid])

    return _pop_fairly(d_lm, size)


def train_sample_leak(l_lm, size):
    d_lm = defaultdict(list)
    for lm in l_lm:
        d_lm[lm.lt.ltid].append(lm)
    for ltid in d_lm:
        random.shuffle(d_lm[ltid])

    return _pop_fairly(d_lm, size)


def make_crf_train(conf, iterobj, return_ltidlist=False):
    method = conf["log_template_crf"]["sample_method"]
    size = conf.getint("log_template_crf", "n_sample")
    if method == "all":
        l_train = list(iterobj)
    elif method == "random":
        l_train = train_sample_random(iterobj, size)
    elif method == "ltgen":
        lt_methods = config.getlist(conf, "log_template_crf",
                                    "sample_lt_methods")
        use_mp = conf.getboolean("log_template_crf", "sample_lt_multiprocess")
        table = lt_common.TemplateTable()
        ltgen = amulog.manager.init_ltgen_methods(conf, table, lt_methods,
                                                  multiprocess=use_mp)
        l_train = train_sample_ltgen(iterobj, size, ltgen)
    elif method == "leak":
        l_train = train_sample_leak(iterobj, size)
    else:
        raise NotImplementedError(
            "Invalid sampling method name {0}".format(method))

    if return_ltidlist:
        train_ltidlist = [lm.lt.ltid for lm in l_train]
        return l_train, train_ltidlist
    else:
        return l_train


def crf_trainfile(conf, iterobj):
    from . import lt_crf
    table = lt_common.TemplateTable()
    ltgen = lt_crf.init_ltgen_crf(conf, table)
    l_train = make_crf_train(conf, iterobj)
    l_buf = [_items.items2str(ltgen.trainitems(lm)) for lm in l_train]
    return "\n\n".join(l_buf)


def make_crf_model(conf, iterobj, output=None,
                   return_sampled_messages=False):
    from . import lt_crf
    table = lt_common.TemplateTable()
    ltgen = lt_crf.init_ltgen_crf(conf, table)

    l_train = make_crf_train(conf, iterobj)
    ltgen.init_trainer()
    model_path = ltgen.train(l_train, output)
    assert os.path.exists(model_path)
    _logger.info("generate crf model {0}".format(model_path))

    if return_sampled_messages:
        return model_path, l_train
    else:
        return model_path


def make_crf_model_from_trainfile(conf, fp, output=None):
    from . import lt_crf
    table = lt_common.TemplateTable()
    ltgen = lt_crf.init_ltgen_crf(conf, table)
    ltgen.init_trainer()
    model_path = ltgen.train_from_file(fp, output)
    assert os.path.exists(model_path)
    _logger.info("generate crf model {0}".format(model_path))
    return model_path



#!/usr/bin/env python
# coding: utf-8

from collections import defaultdict
import numpy as np

from amulog.lt_common import REPLACER


def word_accuracy(tpls_true, tpls_pred):
    n_word = 0
    n_correct_word = 0
    for tpl_true, tpl_pred in zip(tpls_true, tpls_pred):
        for word_true, word_pred in zip(tpl_true, tpl_pred):
            n_word += 1
            if word_true == word_pred:
                n_correct_word += 1
    return 1. * n_correct_word / n_word


def line_accuracy(tpls_true, tpls_pred):
    n_line = 0
    n_correct_line = 0
    for tpl_true, tpl_pred in zip(tpls_true, tpls_pred):
        n_line += 1
        if tpl_true == tpl_pred:
            n_correct_line += 1
    return 1. * n_correct_line / n_line


def tpl_accuracy(tpls_true, tpls_pred, labels_true):
    d_n_line = defaultdict(int)
    d_n_correct_line = defaultdict(int)
    for tpl_true, tpl_pred, cluster in zip(tpls_true, tpls_pred,
                                           labels_true):
        d_n_line[cluster] += 1
        if tpl_true == tpl_pred:
            d_n_correct_line[cluster] += 1

    l_line_accuracy = [1. * d_n_correct_line[label] / d_n_line[label]
                       for label in d_n_line]
    return np.average(l_line_accuracy)


def tpl_word_accuracy(tpls_true, tpls_pred, labels_true):
    d_n_word = defaultdict(int)
    d_n_correct_word = defaultdict(int)
    for tpl_true, tpl_pred, cluster in zip(tpls_true, tpls_pred,
                                           labels_true):
        for word_true, word_pred in zip(tpl_true, tpl_pred):
            d_n_word[cluster] += 1
            if word_true == word_pred:
                d_n_correct_word[cluster] += 1
    l_word_accuracy = [1. * d_n_correct_word[label] / d_n_word[label]
                       for label in d_n_word]
    return np.average(l_word_accuracy)


def tpl_desc_accuracy(tpls_true, tpls_pred, labels_true):
    d_n_desc = defaultdict(int)
    d_n_correct_desc = defaultdict(int)
    for tpl_true, tpl_pred, cluster in zip(tpls_true, tpls_pred,
                                           labels_true):
        for word_true, word_pred in zip(tpl_true, tpl_pred):
            if word_true != REPLACER:
                d_n_desc[cluster] += 1
                if word_true == word_pred:
                    d_n_correct_desc[cluster] += 1
    l_ratio = [1. * d_n_correct_desc[label] / d_n_desc[label]
               for label in d_n_desc]
    return np.average(l_ratio)


def tpl_var_accuracy(tpls_true, tpls_pred, labels_true):
    d_n_var = defaultdict(int)
    d_n_correct_var = defaultdict(int)
    for tpl_true, tpl_pred, cluster in zip(tpls_true, tpls_pred,
                                           labels_true):
        for word_true, word_pred in zip(tpl_true, tpl_pred):
            if word_true == REPLACER:
                d_n_var[cluster] += 1
                if word_true == word_pred:
                    d_n_correct_var[cluster] += 1
    l_ratio = [1. * d_n_correct_var[label] / d_n_var[label]
               for label in d_n_var]
    return np.average(l_ratio)



import queue
import logging
import multiprocessing

_logger = logging.getLogger(__package__)


class WorkerProcess(multiprocessing.Process):

    def __init__(self, task_queue, result_queue, timeout, *super_args, **super_kwargs):
        multiprocessing.Process.__init__(self, *super_args, **super_kwargs)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self._timeout = timeout
        _logger.debug("process {0} ready".format(self._name))

    def run_current_proc(self):
        _logger.info("mproc_queue runs in no multiprocessing mode")
        while True:
            try:
                next_task = self.task_queue.get_nowait()
                if next_task is None:
                     break
                else:
                    ret = self._target(next_task, *self._args, **self._kwargs)
                    self.result_queue.put_nowait(ret)
                    self.task_queue.task_done()
            except queue.Empty:
                break

    def run(self):
        try:
            while True:
                next_task = self.task_queue.get(True, self._timeout)
                if next_task is None:
                    _logger.debug("process {0} exiting".format(self._name))
                    self.task_queue.task_done()
                    break
                else:
                    _logger.debug("process {0} start next task".format(self._name))
                    ret = self._target(next_task, *self._args, **self._kwargs)
                    self.result_queue.put_nowait(ret)
                    _logger.debug("process {0} ended a task".format(self._name))
                    self.task_queue.task_done()
        except queue.Empty:
            _logger.debug("process {0} timeout".format(self._name))
        finally:
            return


class Manager(object):

    def __init__(self, target, n_proc, args=(), kwargs=None, namer=None,
                 timeout=3600):
        if kwargs is None:
            kwargs = {}
        if namer is None:
            namer = lambda x: None
        self._n_proc = n_proc
        self.task_queue = multiprocessing.JoinableQueue()
        self.result_queue = multiprocessing.Queue()
        self._l_worker = [WorkerProcess(self.task_queue, self.result_queue,
                                        timeout, target=target, name=namer(i),
                                        args=args, kwargs=kwargs)
                          for i in range(max(n_proc, 1))]
        if self._n_proc > 0:
            for w in self._l_worker:
                w.start()

    def add(self, task):
        self.task_queue.put(task)

    def add_from(self, l_task):
        for task in l_task:
            self.task_queue.put(task)

    def join(self):
        if self._n_proc > 0:
            self.task_queue.join()
        else:
            # debug without multiprocessing
            self._l_worker[0].run_current_proc()

    def get(self, block=False, timeout=1):
        return self.result_queue.get(block, timeout)

    def get_from(self, l_task, block=False, timeout=1):
        for _ in l_task:
            yield self.result_queue.get(block, timeout)

    def get_all(self):
        while True:
            try:
                ret = self.result_queue.get_nowait()
            except queue.Empty:
                break
            else:
                yield ret

    def is_clean(self):
        assert self.task_queue.empty()
        assert self.result_queue.empty()

    def close(self):
        if self._n_proc > 0:
            for _ in self._l_worker:
                self.task_queue.put(None)
            self.task_queue.join()
        else:
            pass





#!/usr/bin/env python
# coding: utf-8

import os
import unittest
import tempfile

from amulog import common
from amulog import config
from amulog import log_db
from amulog import manager

from amulog import testutil


class TestDB(unittest.TestCase):

    _path_testlog = None
    _path_testdb = None
    _path_ltgendump = None

    @classmethod
    def setUpClass(cls):
        fd_testlog, cls._path_testlog = tempfile.mkstemp()
        os.close(fd_testlog)
        fd_testdb, cls._path_testdb = tempfile.mkstemp()
        os.close(fd_testdb)
        fd_ltgendump, cls._path_ltgendump = tempfile.mkstemp()
        os.close(fd_ltgendump)

        cls._conf = config.open_config(verbose=False)
        cls._conf['general']['src_path'] = cls._path_testlog
        cls._conf['database']['sqlite3_filename'] = cls._path_testdb
        cls._conf['manager']['indata_filename'] = cls._path_ltgendump

        tlg = testutil.TestLogGenerator(testutil.DEFAULT_CONFIG, seed=3)
        tlg.dump_log(cls._path_testlog)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls._path_testlog)
        os.remove(cls._path_testdb)
        os.remove(cls._path_ltgendump)

    def test_makedb_online(self):
        from amulog import __main__ as amulog_main
        targets = amulog_main.get_targets_conf(self._conf)
        manager.process_files_online(self._conf, targets, reset_db=True)

        ld = log_db.LogData(self._conf)
        num = ld.count_lines()
        self.assertEqual(num, 6539,
                         "not all logs added to database")
        ltg_num = len([gid for gid in ld.iter_ltgid()])
        self.assertTrue(3 < ltg_num < 20,
                        ("log template generation fails? "
                         "(groups: {0})".format(ltg_num)))

    def test_makedb_offline(self):
        from amulog import __main__ as amulog_main
        targets = amulog_main.get_targets_conf(self._conf)
        manager.process_files_offline(self._conf, targets, reset_db=True)

        ld = log_db.LogData(self._conf)
        num = ld.count_lines()
        self.assertEqual(num, 6539,
                         "not all logs added to database")
        ltg_num = len([gid for gid in ld.iter_ltgid()])
        self.assertTrue(3 < ltg_num < 20,
                        ("log template generation fails? "
                         "(groups: {0})".format(ltg_num)))

    def test_makedb_parallel(self):
        import copy
        conf = copy.copy(self._conf)
        conf["manager"]["n_process"] = "2"
        conf["log_template"]["lt_methods"] = "re"
        conf["log_template_re"]["variable_rule"] = \
            common.filepath_local(__file__, "test_re.conf")

        from amulog import __main__ as amulog_main
        targets = amulog_main.get_targets_conf(conf)
        manager.process_files_offline(conf, targets, reset_db=True, parallel=True)

        ld = log_db.LogData(self._conf)
        #for ltobj in ld.iter_lt():
        #    print(ltobj)
        num = ld.count_lines()
        self.assertEqual(num, 6539,
                         "not all logs added to database")
        ltg_num = len([gid for gid in ld.iter_ltgid()])
        self.assertTrue(3 < ltg_num < 20,
                        ("log template generation fails? "
                         "(groups: {0})".format(ltg_num)))

    def test_anonymize_overwrite(self):
        from amulog import __main__ as amulog_main
        targets = amulog_main.get_targets_conf(self._conf)
        manager.process_files_online(self._conf, targets, reset_db=True)

        from amulog import anonymize
        am = anonymize.AnonymizeMapper(self._conf)
        am.anonymize()

        import re
        reobj_message = re.compile(r"^[ *#]*$")
        reobj_host = re.compile(r"^host\d+$")
        ld = log_db.LogData(self._conf)
        for lm in ld.iter_lines(ltid=0):
            message = lm.restore_message()
            self.assertTrue(reobj_message.match(message))
            self.assertTrue(reobj_host.match(lm.host))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# coding: utf-8

import re
import logging
from collections import defaultdict
from abc import ABC, abstractmethod

from . import strutil

REPLACER = "**"
REPLACER_HEAD = "*"
REPLACER_TAIL = "*"
REPLACER_REGEX = re.compile(r"\*[A-Z]*?\*")  # shortest match
ANONYMIZED_DESC = "##"

_logger = logging.getLogger(__package__)


class LTTable:

    def __init__(self):
        self._ltdict = {}

    def __iter__(self):
        return self._generator()

    def _generator(self):
        for ltid in self._ltdict:
            yield self._ltdict[ltid]

    def __len__(self):
        return len(self._ltdict)

    def __getitem__(self, key):
        assert isinstance(key, int)
        if key not in self._ltdict:
            raise IndexError("index out of range")
        return self._ltdict[key]

    def next_ltid(self):
        cnt = 0
        while cnt in self._ltdict:
            cnt += 1
        else:
            return cnt

    def restore_lt(self, ltid, ltgid, ltw, lts, count):
        assert ltid not in self._ltdict
        self._ltdict[ltid] = LogTemplate(ltid, ltgid, ltw, lts, count)

    def add_lt(self, ltline):
        assert ltline.ltid not in self._ltdict
        self._ltdict[ltline.ltid] = ltline

    def update_lt(self, ltobj):
        assert ltobj.ltid in self._ltdict
        self._ltdict[ltobj.ltid] = ltobj

    def remove_lt(self, ltid):
        self._ltdict.pop(ltid)


class LogTemplate:

    def __init__(self, ltid, ltgid, ltw, lts, count):
        if len(ltw) == 0:
            raise ValueError("empty ltw, failed to generate LogTemplate")
        self.ltid = ltid
        self.ltgid = ltgid
        self.ltw = ltw
        self.lts = lts
        self.count = count

    def __iter__(self):
        return self.ltw

    def __str__(self):
        return self.restore_message(self.ltw)

    def get(self, key):
        if key == "ltid":
            return self.ltid
        elif key == "ltgid":
            return self.ltgid
        else:
            raise KeyError

    def var(self, l_w):
        if len(l_w) == 0:
            return [REPLACER for w in self.ltw if w == REPLACER]
        else:
            return [w_org for w_org, w_lt in zip(l_w, self.ltw)
                    if w_lt == REPLACER]

    def var_location(self):
        return [i for i, w_lt in enumerate(self.ltw) if w_lt == REPLACER]

    def restore_message(self, l_w, esc=False):
        if l_w is None or len(l_w) == 0:
            l_w = self.ltw
        if esc:
            l_w = [w for w in l_w]
        else:
            l_w = [strutil.restore_esc(w) for w in l_w]

        if self.lts is None:
            return "".join(l_w)
        else:
            return "".join([s + w for w, s in zip(l_w + [""], self.lts)])

    def increment(self):
        self.count += 1
        return self.count

    def replace(self, l_w, l_s=None, count=None):
        self.ltw = l_w
        if l_s is not None:
            self.lts = l_s
        if count is not None:
            self.count = count


class TemplateTable:
    """Temporal template table for log template generator."""

    def __init__(self):
        self._d_tpl = {}  # key = tid, val = template
        self._d_rtpl = {}  # key = key_template, val = tid
        self._d_ltid = {}  # key = tid, val = ltid
        self._d_cand = defaultdict(list)  # key = tid, val = List[ltid]
        self._last_modified = None  # used for LTGenJoint

    def __str__(self):
        ret = []
        for tid, tpl in self._d_tpl.items():
            ret.append(" ".join([str(tid)] + tpl))
        return "\n".join(ret)

    def __iter__(self):
        return self._generator()

    def _generator(self):
        for tid in self._d_tpl:
            yield self._d_tpl[tid]

    def __getitem__(self, key):
        assert isinstance(key, int)
        if key not in self._d_tpl:
            raise IndexError("index out of range")
        return self._d_tpl[key]

    def __len__(self):
        return len(self._d_tpl)

    def next_tid(self):
        cnt = 0
        while cnt in self._d_tpl:
            cnt += 1
        else:
            return cnt

    def tids(self):
        return self._d_tpl.keys()

    @staticmethod
    def _key_template(template):
        # l_word = [strutil.add_esc(w) for w in template]
        # return "@".join(l_word)
        return tuple(template)

    def exists(self, template):
        key = self._key_template(template)
        return key in self._d_rtpl

    def get_tid(self, template):
        key = self._key_template(template)
        return self._d_rtpl[key]

    def get_template(self, tid):
        return self._d_tpl[tid]

    def add(self, template):
        tid = self.next_tid()
        self._d_tpl[tid] = template
        self._d_rtpl[self._key_template(template)] = tid
        return tid

    def replace(self, tid, template):
        self._last_modified = self._d_tpl[tid]
        self._d_tpl[tid] = template
        self._d_rtpl[self._key_template(template)] = tid

    def get_updated(self):
        return self._last_modified

    def add_ltid(self, tid, ltid):
        self._d_ltid[tid] = ltid

    def get_ltid(self, tid):
        return self._d_ltid[tid]

#    def getcand(self, tid):
#        return self._d_cand[tid]
#
#    def addcand(self, tid, ltid):
#        self._d_cand[tid].append(ltid)

    def load(self, obj):
        self._d_tpl, self._d_cand = obj
        for tid, tpl in self._d_tpl.items():
            self._d_rtpl[self._key_template(tpl)] = tid

    def dumpobj(self):
        return self._d_tpl, self._d_cand


class LTGen(ABC):
    state_added = 0
    state_changed = 1
    state_unchanged = 2

    def __init__(self, table):
        if table is None:
            self._table = TemplateTable()
        else:
            self._table = table

    def is_stateful(self):
        return True

    def get_tpl(self, tid):
        return self._table[tid]

    def add_tpl(self, ltw, tid=None):
        if tid is None:
            tid = self._table.add(ltw)
        return tid

    def update_tpl(self, ltw, tid):
        self._table.replace(tid, ltw)

    def merge_tpl(self, l_w, tid):
        old_tpl = self._table[tid]
        new_tpl = merged_template(old_tpl, l_w)
        if old_tpl == new_tpl:
            return self.state_unchanged
        else:
            self.update_tpl(new_tpl, tid)
            return self.state_changed

    def update_table(self, tpl):
        if tpl is None:
            return None, None
        elif self._table.exists(tpl):
            tid = self._table.get_tid(tpl)
            return tid, self.state_unchanged
        else:
            tid = self._table.add(tpl)
            return tid, self.state_added

    def preprocess(self, plines):
        """This function do the pre-process for given data if needed.
        This function is called by process_init_data."""
        pass

    def process_offline(self, d_pline):
        """If there is no need of special process for init phase,
        this function simply call process_line multiple times.
        """
        self.preprocess(d_pline.values())
        d = {}
        for mid, pline in d_pline.items():
            tid, state = self.process_line(pline)
            d[mid] = tid
        return d

    @abstractmethod
    def process_line(self, pline):
        """Estimate log template for given message.
        This method works in incremental processing phase.

        Args:
            pline (dict): parsed log message with log2seq

        Returns:
            tid (int): A template id in TemplateTable.
            state (int)
        """
        raise NotImplementedError

    def postprocess(self, plines):
        """This function do the post-process for given data if needed.
        This function is called by process_init_data."""
        pass

    @abstractmethod
    def load(self, loadobj):
        raise NotImplementedError

    @abstractmethod
    def dumpobj(self):
        raise NotImplementedError


class LTGenOffline(LTGen, ABC):

    @abstractmethod
    def process_offline(self, plines):
        raise NotImplementedError

    def process_line(self, pline):
        msg = "offline LTGen does not support incremental processing"
        raise RuntimeError(msg)

    def load(self, _):
        # offline LTGen does not support suspension and restart
        pass

    def dumpobj(self):
        # offline LTGen does not support suspension and restart
        return None


class LTGenStateless(LTGen, ABC):

    """Subclasses of LTGenStateless is acceptable
    for multiprocessing in offline template generation."""

    @abstractmethod
    def generate_tpl(self, pline):
        raise NotImplementedError

    def is_stateful(self):
        return False

    def process_line(self, pline):
        tpl = self.generate_tpl(pline)
        return self.update_table(tpl)

    def load(self, loadobj):
        # stateless
        pass

    def dumpobj(self):
        # stateless
        return None


class LTGenJoint(LTGen):

    def __init__(self, table: TemplateTable, l_ltgen, ltgen_import_index=None):
        super().__init__(table)
        self._l_ltgen = l_ltgen
        self._import_index = ltgen_import_index

        class_names = [c.__class__.__name__ for c in self._l_ltgen]
        _logger.warning("LTGenJoint init with {0}".format(class_names))

    def is_stateful(self):
        return any([ltgen.is_stateful() for ltgen in self._l_ltgen])

    def _update_ltmap(self, pline, index, tid, state):
        from .lt_import import LTGenImport
        if (self._import_index is not None) and \
                (index != self._import_index):
            ltgen_import: LTGenImport = self._l_ltgen[self._import_index]
            new_tpl = self._table.get_template(tid)
            if state == self.state_added:
                ltgen_import.add_definition(new_tpl)
                msg = "LTGenJoint new template: {0}".format(new_tpl)
                _logger.debug(msg)
            elif state == self.state_changed:
                old_tpl = self._table.get_updated()
                ltgen_import.update_definition(old_tpl, new_tpl)
                msg = "LTGenJoint template update: {0} -> {1}".format(
                    old_tpl, new_tpl)
                _logger.debug(msg)

    def process_line(self, pline):
        _logger.debug("LTGenJoint input: {0}".format(pline["words"]))
        for index, ltgen in enumerate(self._l_ltgen):
            tid, state = ltgen.process_line(pline)
            if tid is not None:
                msg = ("LTGenJoint: method {0} ".format(index) +
                       "successfully generate a template")
                _logger.debug(msg)
                self._update_ltmap(pline, index, tid, state)
                return tid, state
        else:
            msg = "Template not matched/generated: {0}".format(pline["words"])
            _logger.debug(msg)
            return None, None

    def load(self, loadobj):
        for ltgen, ltgen_data in zip(self._l_ltgen, loadobj):
            ltgen.load(ltgen_data)

    def dumpobj(self):
        return [ltgen.dumpobj() for ltgen in self._l_ltgen]


#class LTGenMultiProcess(LTGenOffline):
#
#    _ltgen: LTGenStateless
#
#    def __init__(self, table: TemplateTable, n_proc, kwargs):
#        super().__init__(table)
#        self._n_proc = n_proc
##        self._ltgen = init_ltgen(**kwargs)
##        assert not self._ltgen.is_stateful(), \
##            "multiprocessing is limited to stateless methods"
#        assert(n_proc < 20)
#
#        from multiprocessing import Pool
#        self._pool = Pool(processes=self._n_proc,
#                          initializer=self._pool_init, initargs=(kwargs,))
#
#    @staticmethod
#    def _pool_init(ltgen_kwargs):
#        global _LTGEN_MP_LOCAL
#        _LTGEN_MP_LOCAL = init_ltgen(**ltgen_kwargs)
#        assert not _LTGEN_MP_LOCAL.is_stateful(), \
#            "multiprocessing is limited to stateless methods"
#
#    @staticmethod
#    def _pool_task(args):
#        ret = []
#        for message_id, pline in args:
#            template = _LTGEN_MP_LOCAL.generate_tpl(pline)
#            ret.append((message_id, template))
#        return ret
#
#        #message_id, pline = args
#        #template = _LTGEN_MP_LOCAL.generate_tpl(pline)
#        #return message_id, template
#
#    def _process_offline_pool(self, plines):
#        l_tmp_args = [(mid, pline)
#                      for mid, pline in enumerate(plines)]
#        l_args = np.array_split(l_tmp_args, self._n_proc * 10)
#        try:
#            d_tpl = {}
#            for ret in self._pool.imap_unordered(self._pool_task, l_args):
#                for mid, tpl in ret:
#                    d_tpl[mid] = tpl
#            self._pool.close()
#        except KeyboardInterrupt:
#            self._pool.terminate()
#            exit()
#        else:
#            ret = {}
#            for mid, tpl in d_tpl.items():
#                tid, _ = self.update_table(tpl)
#                ret[mid] = tid
#            return ret
#
#    def process_offline(self, plines):
#        return self._process_offline_pool(plines)


class LTGroup(ABC):

    def __init__(self):
        self._init_dict()

    def _init_dict(self):
        self._d_group = {}  # key : groupid, val : [ltline, ...]
        self._d_rgroup = {}  # key : ltid, val : groupid

    def _next_groupid(self):
        cnt = 0
        while cnt in self._d_group:
            cnt += 1
        else:
            return cnt

    @abstractmethod
    def make(self):
        raise NotImplementedError

    def add_ltid(self, gid, ltline):
        self._d_group.setdefault(gid, []).append(ltline)
        self._d_rgroup[ltline.ltid] = gid

    def restore_ltg(self, db, table):
        for ltid, ltgid in db.iter_ltg_def():
            self._d_group.setdefault(ltgid, []).append(table[ltid])
            self._d_rgroup[ltid] = ltgid

    def load(self, loadobj):
        pass

    def dumpobj(self):
        return None


class LTGroupOnline(LTGroup, ABC):

    def __init__(self, lttable):
        super().__init__()
        self.lttable = lttable

    @abstractmethod
    def add(self, ltline):
        """Returns ltgid"""
        raise NotImplementedError

    def make(self):
        self.remake_all()

    def remake_all(self):
        self._init_dict()
        for ltline in self.lttable:
            ltgid = self.add(ltline)
            ltline.ltgid = ltgid


class LTGroupDummy(LTGroupOnline):
    # This class will work as a dummy
    # (ltgid is always same as ltid)

    def add(self, ltline):
        gid = ltline.ltid
        self.add_ltid(gid, ltline)
        return gid


class LTGroupOffline(LTGroup, ABC):

    def __init__(self, lttable):
        super().__init__()
        self.lttable = lttable

    @abstractmethod
    def make(self):
        raise NotImplementedError


class LTPostProcess(object):

    def __init__(self, conf, table, lttable, l_alg):
        self._table = table
        self._lttable = lttable
        self._rules = []
        for alg in l_alg:
            if alg == "dummy":
                self._rules.append(VariableLabelRule())
            elif alg == "host":
                self._rules.append(VariableLabelHost(conf))
            else:
                raise NotImplementedError

        self.sym_header = REPLACER_HEAD
        self.sym_footer = REPLACER_TAIL

    def _labeled_variable(self, w):
        return "".join((self.sym_header, w, self.sym_footer))

    def replace_variable(self, l_w, tpl, sym):
        ret = []
        for org_w, tpl_w in zip(l_w, tpl):
            if tpl_w == sym:
                for r in self._rules:
                    ww = r.replace_word(org_w)
                    if ww is not None:
                        ret.append(self._labeled_variable(ww))
                        break
                else:
                    ret.append(tpl_w)
            else:
                ret.append(tpl_w)
        return ret

    def search(self, tid, ltw):
        """Search existing candidates of template derivation. Return None
        if no possible candidates found."""
        l_ltid = self._table.getcand(tid)
        for ltid in l_ltid:
            if self._lttable[ltid].ltw == ltw:
                return ltid
        else:
            return None

        # if len(self._table.getcand(tid)) == 0:
        #    return None
        # else:
        #    return self._table.getcand(tid)[0]


class VariableLabelRule(object):

    def __init__(self):
        pass

    def replace_word(self, w):
        """str: If the given word is a member of some host group,
        return the group name. Otherwise, return None."""
        return None


class VariableLabelHost(VariableLabelRule):

    def __init__(self, conf):
        super().__init__()
        from . import host_alias
        self.ha = host_alias.init_hostalias(conf)

    def replace_word(self, w):
        return self.ha.get_group(w)


def merged_template(m1, m2):
    """Return common area of log message (to be log template)"""
    ret = []
    for w1, w2 in zip(m1, m2):
        if w1 == w2:
            ret.append(w1)
        else:
            ret.append(REPLACER)
    return ret
